# -*- perl -*-

# Copyright (c) 2003 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2003-Jan-30 15:27 (EST)
# Function: database testing
#
# $Id: DataBase.pm,v 1.2 2003/02/05 04:25:48 jaw Exp $

package DataBase;
@ISA = qw(Service);
use Encode;
use Socket;
use Fcntl;

# And mark how well the sequel hangs together
#   -- Shakespeare, King Richard III

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [qw(Service MonEl BaseIO)],
    methods => {},
    versn => '3.2',
    html  => 'dbi',
    fields => {
      db::dsn => {
	  descr => 'database dsn in DBI syntax',
	  exmpl => 'dbi:Pg:my_db=mydb;host=dbhost.example.com',
	  attrs => ['config', 'inherit'],
      },
      db::sql => {
	  descr => 'a select statement',
	  attrs => ['config', 'inherit'],
	  exmpl => "select COUNT(*) from mytable where myfield = 'value'",
      },
      db::user => {
	  descr => 'database username',
	  attrs => ['config', 'inherit'],
      },
      db::pass => {
	  descr => 'database password',
	  attrs => ['config', 'inherit'],
      },
      db::pid => {},
      db::rbuffer => {},
	
    },
};

sub probe {
    my $name = shift;

    return [3, \&config] if $name =~ /^DB/;
}

sub config {
    my $me = shift;
    my $cf = shift;
    
    $me->init_from_config( $doc, 'db' );
    $me->{uname} = "DB_$me->{db}{dsn}";

    return $cf->error( 'dsn not specified' )
	unless $me->{db}{dsn};
    return $cf->error( 'user not specified' )
	unless $me->{db}{user};
    return $cf->error( 'sql select not specified' )
	unless $me->{db}{sql};
    
    bless $me if( ref($me) eq 'Service' );
    $me;
}

sub start {
    my $me = shift;
    my( $pid, $fh );
    
    $me->debug( 'start database' );
    $me->Service::start();
    
    $me->{fd} = $fh = BaseIO::anon_fh();
    unless( socketpair($fh, DB, AF_UNIX, SOCK_STREAM, PF_UNSPEC) ){
	my $m = "pipe failed: $!";
	::sysproblem( "DB $m" );
	$me->debug( $m );
	$me->done();
	return;
    }
    $me->baseio_init();

    unless( ($pid = fork) ){
	# child
        BaseIO::closeall();
	close STDIN;  open( STDIN,  "<&DB" );
	close STDOUT; open( STDOUT, ">&DB" );
	# let stderr go where ever argus's stderr goes
	close $fh;
	close DB;
	exec( "$::libdir/dbhelper",
	      $me->{db}{dsn}, $me->{db}{user}, ($me->{db}{pass} || ''), $me->{db}{sql} );
        syswrite( STDERR, "bummer dude, couldn't run dbhelper: $!\n" );
	_exit(-1);
    }
    close DB;
    
    if( !defined($pid) ){
	# fork failed
	my $m = "fork failed: $!";
	::sysproblem( "DB $m" );
	$me->debug( $m );
	return $me->done();
    }

    $me->{db}{pid} = $pid;
    $me->{db}{rbuffer} = '';
    $me->wantread(1);
    $me->wantwrit(0);
    $me->settimeout( $me->{srvc}{timeout} );
}

sub timeout {
    my $me = shift;

    $me->debug( 'timeout' );
    kill 9, $me->{db}{pid} if $me->{db}{pid};
    $me->finish();
}

sub readable {
    my $me = shift;
    my( $fh, $i, $l );

    $fh = $me->{fd};
    $i = sysread $fh, $l, 8192;
    if( $i ){
	$me->debug( "db - read data ($l)" );
	$me->{db}{rbuffer} .= $l;
    }else{
	$me->finish();
    }
}

sub finish {
    my $me = shift;

    $me->down( $me->{db}{rbuffer} ) if $me->{db}{rbuffer} =~ /^bummer/i;
    return $me->generic_test( $me->{db}{rbuffer}, 'DB' );
}

################################################################
sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->Service::about_more($ctl);
    foreach $k (sort keys %{$me->{db}}){
	$v = $me->{db}{$k};
	next unless defined $v;
	if( ref($v) ){
	    $v = '#<REF>';
	}else{
	    $v = encode( $v );
	}
	$ctl->write( "db::$k:\t$v\n" );
    }
}

sub webpage_more {
    my $me = shift;
    my $fh = shift;
    my( $k, $v );
    
    foreach $k (qw(dsn sql)){
	$v = $me->{db}{$k};
	print $fh "<TR><TD>$k</TD><TD>$v</TD></TR>\n" if defined($v);
    }
}

################################################################
Doc::register( $doc );
push @Service::probes, \&probe;

1;

